from __future__ import annotations

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..models import ConfirmationAtLocation, Formulation, Location, Rx
from ..utils import dispense


@method_decorator(login_required, name="dispatch")
class DispenseView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    model_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/dispense.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"

    def get_context_data(self, **kwargs):
        container_count = self.kwargs.get("container_count") or 0
        location_id = self.kwargs.get("location_id")
        formulation_id = self.kwargs.get("formulation_id")
        try:
            location = Location.objects.get(pk=location_id)
        except ObjectDoesNotExist:
            location = None
        try:
            formulation = Formulation.objects.get(pk=formulation_id)
        except ObjectDoesNotExist:
            formulation = None
        kwargs.update(
            item_count=list(range(1, container_count + 1)),
            locations=Location.objects.filter(site__isnull=False),
            formulations=Formulation.objects.all(),
            location=location,
            formulation=formulation,
        )
        return super().get_context_data(**kwargs)

    def get_rx(self, subject_identifier, location, formulation) -> Rx | None:
        try:
            rx = Rx.objects.get(
                registered_subject__subject_identifier=subject_identifier,
                medications__in=[formulation.medication],
                site=getattr(location, "site", None),
            )
        except ObjectDoesNotExist:
            rx = None
            messages.add_message(
                self.request,
                messages.WARNING,
                f"Subject {subject_identifier} not found at {location.display_name}.",
            )
        else:
            if rx.rx_expiration_date and rx.rx_expiration_date < timezone.now():
                messages.add_message(
                    self.request,
                    messages.WARNING,
                    (
                        f"The prescription for {formulation} for "
                        f"subject {subject_identifier} is expired."
                    ),
                )
                rx = None
        return rx

    @property
    def confirm_at_location(self):
        confirm_at_location_id = self.kwargs.get("confirm_at_location")
        try:
            confirm_at_location = ConfirmationAtLocation.objects.get(id=confirm_at_location_id)
        except ObjectDoesNotExist:
            confirm_at_location = None
            messages.add_message(
                self.request, messages.ERROR, "Invalid stock transfer confirmation."
            )
        return confirm_at_location

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        location_id = request.POST.get("location_id")
        location = Location.objects.get(pk=location_id)
        formulation_id = request.POST.get("formulation_id")
        formulation = Formulation.objects.get(pk=formulation_id)
        subject_identifier = request.POST.get("subject_identifier")
        container_count = request.POST.get("container_count")
        stock_codes = request.POST.getlist("codes") if request.POST.get("codes") else None
        rx = self.get_rx(subject_identifier, location, formulation)

        if location and formulation and rx and container_count:
            if stock_codes:
                dispense_qs = dispense(
                    stock_codes,
                    location,
                    rx,
                    request.user.username,
                    request,
                )
                if dispense_qs:
                    messages.add_message(
                        self.request,
                        messages.SUCCESS,
                        f"Dispensed {dispense_qs.count()} item.",
                    )
                url = reverse(
                    "edc_pharmacy:dispense_url",
                    kwargs={
                        "location_id": location.id,
                        "formulation_id": formulation.id,
                        "subject_identifier": subject_identifier,
                        "container_count": container_count,
                    },
                )
            else:
                url = reverse(
                    "edc_pharmacy:dispense_url",
                    kwargs={
                        "location_id": location.id,
                        "formulation_id": formulation.id,
                        "subject_identifier": subject_identifier,
                        "container_count": container_count,
                    },
                )
            return HttpResponseRedirect(url)
        return HttpResponseRedirect(reverse("edc_pharmacy:home_url"))



from django.apps import apps as django_apps


class TimepointClosed(Exception):  # noqa: N818
    pass


class Timepoint:
    """A class that represents a timepoint model and the relevant
    attributes and values of that model.

    Note: typically `model` is `edc_appointment.appointment`.
    """

    def __init__(self, model=None, datetime_field=None, status_field=None, closed_status=None):
        self.model = model
        self.datetime_field = datetime_field
        self.status_field = status_field
        self.closed_status = closed_status

    def __str__(self):
        return self.model

    @property
    def model_cls(self):
        return django_apps.get_model(self.model)



from django.contrib.auth.models import Group, User
from django.contrib.sites.models import Site
from django.core.validators import RegexValidator
from django.db import models
from django.db.models.deletion import CASCADE
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from edc_export.choices import EXPORT_FORMATS
from edc_export.constants import CSV
from edc_notification.model_mixins import NotificationUserProfileModelMixin

from ..constants import CUSTOM_ROLE, STAFF_ROLE
from .role import Role


class UserProfileError(Exception):
    pass


class UserProfile(NotificationUserProfileModelMixin, models.Model):
    id = models.AutoField(primary_key=True)

    user = models.OneToOneField(User, on_delete=CASCADE)

    is_multisite_viewer = models.BooleanField(
        verbose_name="Multisite viewer",
        default=False,
        help_text="Designates if this user may view data from sites other than the current.",
    )

    sites = models.ManyToManyField(Site, blank=True)

    job_title = models.CharField(max_length=100, default="", blank=True)

    alternate_email = models.EmailField(_("Alternate email address"), blank=True, default="")

    mobile = models.CharField(
        max_length=25,
        validators=[RegexValidator(regex=r"^\+\d+")],
        default="",
        blank=True,
        help_text="e.g. +1234567890",
    )

    clinic_label_printer = models.CharField(
        max_length=100,
        default="",
        blank=True,
        help_text=format_html(
            'Change in <a href="{href}">{label}</a>',
            href="/edc_label/",
            label="Edc Label Administration",
        ),
    )

    lab_label_printer = models.CharField(
        max_length=100,
        default="",
        blank=True,
        help_text=format_html(
            'Change in <a href="{href}">{label}</a>',
            href="/edc_label/",
            label="Edc Label Administration",
        ),
    )

    print_server = models.CharField(
        max_length=100,
        default="",
        blank=True,
        help_text=format_html(
            'Change in <a href="{href}">{label}</a>',
            href="/edc_label/",
            label="Edc Label Administration",
        ),
    )

    export_format = models.CharField(
        verbose_name="Export format",
        max_length=25,
        choices=EXPORT_FORMATS,
        default=CSV,
        blank=True,
        help_text="Note: requires export permissions",
    )

    roles = models.ManyToManyField(Role, blank=True)

    def __str__(self):
        return self.user.username

    def add_groups_for_roles(self, pk_set):  # noqa: ARG002
        """Add groups to this user for the selected roles.

        Called by m2m signal.
        """
        if CUSTOM_ROLE not in [obj.name for obj in self.roles.all()]:
            group_names = [group.name for group in self.user.groups.all()]
            add_group_names = []
            for role in self.roles.all():
                add_group_names.extend(
                    [
                        group.name
                        for group in role.groups.all()
                        if group.name not in group_names
                    ]
                )
            add_group_names = list(set(add_group_names))
            for name in add_group_names:
                self.user.groups.add(Group.objects.get(name=name))
                self.user.save()

    def remove_groups_for_roles(self, pk_set):
        """Remove groups from this user for the removed roles.

        Called by m2m signal.
        """
        if CUSTOM_ROLE in [obj.name for obj in Role.objects.filter(pk__in=pk_set)]:
            self.user.groups.clear()
            self.user.userprofile.roles.clear()
            self.user.userprofile.roles.add(Role.objects.get(name=STAFF_ROLE))
        else:
            remove_group_names = []
            current_group_names = []
            for role in self.roles.all():
                current_group_names.extend([group.name for group in role.groups.all()])
            for role in Role.objects.filter(pk__in=pk_set):
                remove_group_names.extend(
                    [
                        group.name
                        for group in role.groups.all()
                        if group.name not in current_group_names
                    ]
                )
            for name in remove_group_names:
                self.user.groups.remove(Group.objects.get(name=name))

    class Meta(NotificationUserProfileModelMixin.Meta):
        verbose_name = _("User profile")
        verbose_name_plural = _("User profiles")



from math import floor
from typing import Any

from django.apps import apps as django_apps
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch
from django.views.generic.base import ContextMixin


class AdministrationDashboardError(Exception):
    pass


class AdministrationViewMixin(ContextMixin):
    template_name = "edc_dashboard/administration.html"

    edc_module_prefix = "Edc"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        sections = self.sections
        kwargs.update(sections=sections)
        index = 0
        col_one = {}
        col_two = {}
        col_three = {}
        per_col = floor(len(list(sections.keys())) / 3)
        for k, v in sections.items():
            index += 1
            if index <= per_col:
                col_one.update({k: v})
            elif per_col < index < (per_col * 2) + 1:
                col_two.update({k: v})
            else:
                col_three.update({k: v})
        kwargs.update(col_one=col_one, col_two=col_two, col_three=col_three)
        return super().get_context_data(**kwargs)

    @staticmethod
    def get_section(app_config=None):
        """Returns a dictionary for a single section.

        Format is {verbose_name: url_name}

        To be included must be a registered namespace. To main apps
        urls.py add paths. For example for `ambition_export`:

            urlpatterns = [
            ...
            path('admin/', ambition_export_admin.urls),
            path('ambition_export/', include('ambition_export.urls')),
            ...
            ]

        """
        section = {}
        try:
            url_namespace = app_config.url_namespace
        except AttributeError:
            url_namespace = app_config.name
        try:
            url = app_config.home_url_name
        except AttributeError:
            url = f"{url_namespace}:home_url"
        try:
            reverse(url)
        except NoReverseMatch as e:
            raise AdministrationDashboardError(
                f"Administrative section `{app_config.name}` url name is invalid. "
                f"Either fix the url name or set "
                f"`include_in_administration_section=False` in the AppConfig. "
                f"url_name=`{url}`. Got {e}"
            )
            # probably is not a registered namespace
            pass
        else:
            section = {app_config.verbose_name: url}
        return section

    @property
    def default_sections(self):
        return {
            "Accounts, Profiles, Logs": "edc_auth:home_url",
            # "Sites": "edc_sites_admin:home_url",
        }

    @property
    def sections(self):
        """Returns a dictionary (sorted) of the administration sections
        to show on the Administration page.
        """
        sections = {}
        for app_config in django_apps.get_app_configs():
            try:
                include = app_config.include_in_administration_section
            except AttributeError:
                include = False
            if include:
                sections.update(**self.get_section(app_config))
        sections.update(**self.default_sections)

        edc_sections = {
            k: v for k, v in sections.items() if k.startswith(self.edc_module_prefix)
        }
        edc_keys = list(edc_sections.keys())
        edc_keys.sort()

        other_sections = {
            k: v for k, v in sections.items() if not k.startswith(self.edc_module_prefix)
        }
        other_keys = list(other_sections.keys())
        other_keys.sort()

        keys = other_keys + edc_keys
        return {key: sections.get(key) for key in keys}



# Generated by Django 5.1.2 on 2024-11-16 21:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0031_historicalrepackrequest_task_id_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstockrequest",
            name="task_id",
            field=models.UUIDField(null=True),
        ),
        migrations.AddField(
            model_name="stockrequest",
            name="task_id",
            field=models.UUIDField(null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2022-08-24 21:11

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_auth", "0026_rename_permissions_edcpermissions"),
        ("edc_adverse_event", "0006_auto_20210425_1628"),
    ]

    operations = [
        migrations.CreateModel(
            name="EdcPermissions",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_auth.edcpermissions",),
        ),
    ]



from celery.result import AsyncResult
from celery.states import SUCCESS
from clinicedc_constants import CANCEL, CANCELLED, COMPLETE, NULL_STRING, PENDING
from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django_audit_fields import audit_fieldset_tuple
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local
from rangefilter.filters import DateRangeFilterBuilder

from ...admin_site import edc_pharmacy_admin
from ...forms import StockRequestForm
from ...models import StockRequest
from ..actions import allocate_stock_to_subject, prepare_stock_request_items_action
from ..actions.print_labels import print_labels_from_stock_request_by_code
from ..model_admin_mixin import ModelAdminMixin
from ..utils import stock_request_status_counts


class StatusListFilter(admin.SimpleListFilter):
    title = _("Status")
    parameter_name = "status"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (
            (PENDING, _("Pending")),
            (COMPLETE, _("Complete")),
            (CANCELLED, _("Cancelled")),
            ("ZERO", _("Zero total")),
        )

    def queryset(self, request, queryset):  # noqa: ARG002
        if self.value():
            if self.value() == PENDING:
                return (
                    queryset.filter(
                        stockrequestitem__allocation__isnull=True, cancel=NULL_STRING
                    )
                    .exclude(item_count=0)
                    .distinct()
                )
            if self.value() == COMPLETE:
                return (
                    queryset.filter(cancel=NULL_STRING)
                    .exclude(stockrequestitem__allocation__isnull=True)
                    .exclude(item_count=0)
                    .distinct()
                )
            if self.value() == CANCELLED:
                return queryset.exclude(cancel=NULL_STRING)
            if self.value() == "ZERO":
                return queryset.filter(item_count=0, cancel=NULL_STRING)
        return None


@admin.register(StockRequest, site=edc_pharmacy_admin)
class StockRequestAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Request for stock"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20
    ordering = ("-request_identifier",)

    autocomplete_fields = ("container", "formulation", "location")
    form = StockRequestForm

    actions = (
        prepare_stock_request_items_action,
        allocate_stock_to_subject,
        print_labels_from_stock_request_by_code,
        "delete_selected",
    )

    fieldsets = (
        (
            "Section A",
            {
                "fields": (
                    "request_identifier",
                    "request_datetime",
                    "start_datetime",
                    "cutoff_datetime",
                    "location",
                )
            },
        ),
        (
            "Section B",
            {"fields": ("formulation", "container", "containers_per_subject")},
        ),
        (
            "Section C",
            {"fields": ("item_count",)},
        ),
        (
            "Section D: Customize this request",
            {"fields": ("subject_identifiers", "excluded_subject_identifiers")},
        ),
        (
            "Section E: Cancel this request",
            {
                "description": (
                    "A request may only be cancelled before stock is allocated by "
                    "the central pharmacy. The EDC will check if the request may be cancelled."
                ),
                "fields": ("cancel",),
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "stock_request_id",
        "stock_request_date",
        "stock_start_date",
        "stock_cutoff_date",
        "requested_from",
        "product_column",
        "stock_request_status",
        "links_changelist",
        "task_status",
    )

    list_filter = (
        "location",
        ("request_datetime", DateRangeFilterBuilder()),
        StatusListFilter,
        "formulation",
        "container",
        "start_datetime",
        "cutoff_datetime",
    )

    search_fields = (
        "id",
        "request_identifier",
        "stockrequestitem__allocation__id",
        "stockrequestitem__allocation__registered_subject__subject_identifier",
        "stockrequestitem__allocation__stock__id",
        "stockrequestitem__allocation__stock__code",
    )

    readonly_fields = ("item_count",)

    def redirect_url(self, request, obj, post_url_continue=None) -> str | None:
        """Redirect to the review page immediately after saving model."""
        redirect_url = super().redirect_url(request, obj, post_url_continue)
        if obj.cancel == CANCEL:
            pass
        elif not obj.stockrequestitem_set.all().exists():
            redirect_url = reverse(
                "edc_pharmacy:review_stock_request_url",
                kwargs={"stock_request": obj.pk},
            )
        return redirect_url

    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        if obj and obj.stockrequestitem_set.all().exists():
            fields = (
                "request_identifier",
                "request_datetime",
                "start_datetime",
                "cutoff_datetime",
                "location",
                "formulation",
                "container",
                "containers_per_subject",
                "item_count",
                "subject_identifiers",
                # "excluded_subject_identifiers",
            )
        return fields

    @admin.display(description="Request #", ordering="request_identifier")
    def stock_request_id(self, obj):
        return obj.request_identifier

    @admin.display(description="PER", ordering="containers_per_subject")
    def per_subject(self, obj):
        return obj.containers_per_subject

    @admin.display(description="From", ordering="location")
    def requested_from(self, obj):
        return obj.location

    @admin.display(description="items", ordering="item_count")
    def request_item_count(self, obj):
        return obj.item_count

    @admin.display(description="Container", ordering="container_name")
    def container_str(self, obj):
        return mark_safe("<BR>".join(str(obj.container).split(" ")))  # noqa: S308

    @admin.display(description="Task")
    def task_status(self, obj):
        if obj.task_id:
            result = AsyncResult(str(obj.task_id))
            return getattr(result, "status", None)
        return None

    @admin.display(description="Product")
    def product_column(self, obj):
        context = dict(
            formulation=obj.formulation,
            containers_per_subject=obj.containers_per_subject,
            container=obj.container,
        )
        return render_to_string(
            "edc_pharmacy/stock/stock_request_product_column.html", context=context
        )

    @admin.display(description="Links")
    def links_changelist(self, obj):
        links = []
        # stockrequestitems
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockrequestitem_changelist")
        url = f"{url}?q={obj.request_identifier}"
        context = dict(url=url, label="Requested", title="Go to stock request items")
        links.append(context)
        # allocations
        url = reverse("edc_pharmacy_admin:edc_pharmacy_allocation_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(
            url=url,
            label="Allocations",
            title="Go to allocation",
        )
        links.append(context)
        # stock
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        links.append(context)
        return render_to_string(
            "edc_pharmacy/stock/stock_request_links_column.html",
            context={"links": links},
        )

    @admin.display(description="Status")
    def stock_request_status(self, obj):
        if obj.cancel == CANCEL:
            return "CANCELLED"
        context = stock_request_status_counts(obj)
        context.update(task_status=self.task_status(obj), success=SUCCESS)
        return render_to_string(
            "edc_pharmacy/stock/stock_request_status_column.html",
            context=context,
        )

    @admin.display(description="Request date", ordering="request_datetime")
    def stock_request_date(self, obj):
        return to_local(obj.request_datetime).date()

    @admin.display(description="Start", ordering="start_datetime")
    def stock_start_date(self, obj):
        if obj and obj.start_datetime:
            return to_local(obj.start_datetime).date()
        return None

    @admin.display(description="Cutoff", ordering="cutoff_datetime")
    def stock_cutoff_date(self, obj):
        if obj and obj.cutoff_datetime:
            return to_local(obj.cutoff_datetime).date()
        return None



from django.contrib import admin
from django.contrib.admin.decorators import register
from django.contrib.sites.shortcuts import get_current_site

from edc_auth.admin import UserAdmin as BaseUserAdmin
from edc_lab.admin import PanelAdmin as BasePanelAdmin
from edc_model_admin.mixins import TemplatesModelAdminMixin
from edc_sites.admin import SiteModelAdminMixin
from edc_visit_schedule.admin import VisitScheduleAdmin as BaseVisitScheduleAdmin

from ..admin_site import edc_data_manager_admin
from ..auth_objects import DATA_MANAGER
from ..models import (
    CrfDataDictionary,
    DataManagerUser,
    QuerySubject,
    QueryUser,
    QueryVisitSchedule,
    RequisitionDataDictionary,
    RequisitionPanel,
    VisitDataDictionary,
)


@register(VisitDataDictionary, site=edc_data_manager_admin)
class VisitDataDictionaryAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    ordering = ("model",)
    search_fields = ("model",)


@register(CrfDataDictionary, site=edc_data_manager_admin)
class CrfDataDictionaryAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    ordering = ("model", "number")
    search_fields = ("model", "prompt", "number")


@register(RequisitionDataDictionary, site=edc_data_manager_admin)
class RequisitionDataDictionaryAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    ordering = ("model", "number")
    search_fields = ("model", "prompt", "number")


@register(QuerySubject, site=edc_data_manager_admin)
class QuerySubjectAdmin(SiteModelAdminMixin, admin.ModelAdmin):
    ordering = ("subject_identifier",)
    search_fields = ("subject_identifier",)


@register(RequisitionPanel, site=edc_data_manager_admin)
class RequisitionPanelAdmin(TemplatesModelAdminMixin, BasePanelAdmin):
    ordering = ("display_name",)
    search_fields = ("display_name", "name")


@register(DataManagerUser, site=edc_data_manager_admin)
class DataManagerUserAdmin(BaseUserAdmin):
    ordering = ("first_name", "last_name")
    search_fields = ("first_name", "last_name", "username", "email")

    inlines = []

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .filter(
                groups__name=DATA_MANAGER,
                userprofile__sites__id=get_current_site(request).id,
            )
        )


@register(QueryUser, site=edc_data_manager_admin)
class QueryUserAdmin(BaseUserAdmin):
    ordering = ("first_name", "last_name")
    search_fields = ("first_name", "last_name", "username", "email")

    inlines = []

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .filter(userprofile__sites__id=get_current_site(request).id)
            .exclude(groups__name=DATA_MANAGER)
        )


@register(QueryVisitSchedule, site=edc_data_manager_admin)
class QueryVisitScheduleAdmin(TemplatesModelAdminMixin, BaseVisitScheduleAdmin):
    pass



from __future__ import annotations

from typing import Any

from clinicedc_constants import HIGH_PRIORITY
from django.apps import apps as django_apps
from django.db import models

from edc_model.models import BaseUuidModel

from ..choices import PRIORITY
from ..exceptions import ActionTypeError

MISSING_ACTION_CLASS = "Model missing an action class. See {reference_model_cls!r}"
NOT_CONFIGURED_FOR_ACTIONS = (
    "Model not configured for Actions. Are you using the model mixin? "
    "See {reference_model_cls!r}. Got {err}"
)
REFERENCE_MODEL_LOOKUP_ERROR = "{err}. Got reference_model=`{reference_model}`"


class ActionTypeManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name=name)


class ActionType(BaseUuidModel):
    name = models.CharField(max_length=50, unique=True)

    display_name = models.CharField(max_length=100)

    reference_model = models.CharField(
        max_length=100,
        blank=True,
        default="",
        help_text="reference model",
    )

    related_reference_model = models.CharField(
        max_length=100,
        blank=True,
        default="",
        help_text="origin reference model",
    )

    priority = models.CharField(max_length=25, choices=PRIORITY, default=HIGH_PRIORITY)

    show_on_dashboard = models.BooleanField(default=False)

    show_link_to_changelist = models.BooleanField(default=False)

    create_by_action = models.BooleanField(
        default=True, help_text="This action may be created by another action"
    )

    create_by_user = models.BooleanField(
        default=True, help_text="This action may be created by the user"
    )

    instructions = models.TextField(max_length=250, blank=True, default="")

    objects = ActionTypeManager()

    def __str__(self) -> str:
        return self.display_name

    def natural_key(self) -> tuple:
        return (self.name,)

    @property
    def reference_model_cls(self) -> Any:
        model_cls = None
        if self.reference_model:
            try:
                model_cls = django_apps.get_model(self.reference_model)
            except (LookupError, ValueError) as e:
                raise ActionTypeError(
                    REFERENCE_MODEL_LOOKUP_ERROR.format(
                        err=str(e), reference_model=self.reference_model
                    )
                ) from e
        return model_cls

    def save(self, *args, **kwargs):
        self.display_name = self.display_name or self.name
        if self.reference_model and self.reference_model != "edc_action_item.reference":
            try:
                if not self.reference_model_cls.action_name:
                    raise ActionTypeError(
                        MISSING_ACTION_CLASS.format(
                            reference_model_cls=self.reference_model_cls
                        )
                    )
            except AttributeError as e:
                if "action_name" in str(e):
                    raise ActionTypeError(
                        NOT_CONFIGURED_FOR_ACTIONS.format(
                            reference_model_cls=self.reference_model_cls, err=str(e)
                        )
                    ) from e
                raise
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        indexes = (
            *BaseUuidModel.Meta.indexes,
            models.Index(fields=["id", "name"]),
        )
        default_permissions = ("add", "change", "delete", "view", "export", "import")



from django import forms

from edc_form_validators import FormValidator, FormValidatorMixin
from edc_model_form.mixins import BaseModelFormMixin
from edc_sites.modelform_mixins import SiteModelFormMixin

from ..models import Note


class NoteFormValidator(FormValidator):
    def clean(self):
        self.required_if_true(True, field_required="note")


class NoteForm(
    SiteModelFormMixin,
    BaseModelFormMixin,
    FormValidatorMixin,
    forms.ModelForm,
):
    report_datetime_field_attr = "report_datetime"
    form_validator_cls = NoteFormValidator

    class Meta:
        model = Note
        fields = "__all__"
        help_text = {"subject_identifier": "(read-only)", "name": "(read-only)"}  # noqa: RUF012
        widgets = {  # noqa: RUF012
            "report_model": forms.TextInput(attrs={"readonly": "readonly"}),
            "subject_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "name": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from __future__ import annotations

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction

from ..managers import CurrentSiteManager
from ..site import sites
from ..utils import get_site_model_cls


class SiteModelMixinError(Exception):
    pass


class SiteModelMixin(models.Model):
    site = models.ForeignKey(
        "sites.site",
        on_delete=models.PROTECT,
        null=True,
        related_name="+",
    )

    on_site = CurrentSiteManager()

    def save(self, *args, **kwargs):
        self.update_site_on_save(*args, **kwargs)
        super().save(*args, **kwargs)

    def update_site_on_save(self, *args, **kwargs) -> None:  # noqa: ARG002
        if not self.id:
            if not self.site_id and not self.site:
                self.site = self.get_site_on_create()
        elif "update_fields" in kwargs and "site" not in kwargs.get("update_fields"):
            pass
        else:
            self.validate_site_against_current()

    def get_site_on_create(self) -> Site:
        """Returns a site model instance.

        See also django-multisite.
        """
        try:
            site_obj = self.site
        except ObjectDoesNotExist:
            site_obj = None
        if not site_obj:
            if self.site_id:
                self.site = Site.objects.get(id=self.site_id)
            else:
                try:
                    with transaction.atomic():
                        site_obj = get_site_model_cls().objects.get_current()
                except ObjectDoesNotExist as e:
                    site_ids = [str(s) for s in sites.all()]
                    raise SiteModelMixinError(
                        "Exception raised when trying manager method `get_current()`. "
                        f"Sites registered with `sites` global are {site_ids}. "
                        f"settings.SITE_ID={settings.SITE_ID}. Got {e}."
                    ) from e
        return site_obj

    def validate_site_against_current(self) -> None:
        """Validate existing site instance matches current_site."""
        return

    class Meta:
        abstract = True



class HolidayError(Exception):
    pass


class HolidayImportError(Exception):
    pass


class HolidayFileNotFoundError(Exception):
    pass


class FacilityError(Exception):
    pass


class FacilitySiteError(Exception):
    pass


class FacilityCountryError(Exception):
    pass



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...forms import StorageBinForm
from ...models import StorageBin
from ..actions import add_to_storage_bin_action
from ..actions.storage_bin import move_to_storage_bin_action
from ..model_admin_mixin import ModelAdminMixin


@admin.register(StorageBin, site=edc_pharmacy_admin)
class StorageBinAdmin(SiteModelAdminMixin, ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Storage Bin"
    change_form_title = "Pharmacy: Storage Bin"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    form = StorageBinForm

    actions = (add_to_storage_bin_action, move_to_storage_bin_action)

    fieldsets = (
        (
            "Section A: Identifier and date",
            {
                "fields": (
                    "bin_identifier",
                    "bin_datetime",
                )
            },
        ),
        (
            "Section B: Storage Bin",
            {
                "fields": (
                    "name",
                    "location",
                    "capacity",
                    "in_use",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "name",
        "location",
        "capacity",
        "bin_date",
        "in_use",
        "storage_bin_item_changelist",
    )

    search_fields = (
        "id",
        "bin_identifier",
        "name",
        "storagebinitem__stock__code",
        "storagebinitem__stock__allocation__registered_subject__subject_identifier",
    )

    list_filter = (
        "in_use",
        "location__display_name",
        ("bin_datetime", DateRangeFilterBuilder()),
    )

    @admin.display(description="Bin #", ordering="-bin_identifier")
    def identifier(self, obj):
        return obj.bin_identifier

    @admin.display(description="Bin date", ordering="bin_datetime")
    def bin_date(self, obj):
        return to_local(obj.bin_datetime).date()

    @admin.display(description="Stock")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Items")
    def storage_bin_item_changelist(self, obj):
        items = obj.storagebinitem_set.all().count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebinitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label=f"Items ({items})", title="Go to bin items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



from edc_model_admin.admin_site import EdcAdminSite

from .apps import AppConfig

__all__ = ["edc_form_runners_admin"]

edc_form_runners_admin = EdcAdminSite(
    name="edc_form_runners_admin", app_label=AppConfig.name, keep_delete_action=True
)



from __future__ import annotations

import sys
from copy import deepcopy
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import import_module, module_has_submodule

if TYPE_CHECKING:
    from edc_form_runners.form_runner import FormRunner


class AlreadyRegistered(Exception):
    pass


class SiteFormRunnerError(Exception):
    pass


__all__ = ["site_form_runners"]


class SiteFormRunners:
    def __init__(self):
        self.registry: dict[str, type[FormRunner]] = {}
        self.loaded = False

    def register(self, runner=None):
        if "makemigrations" not in sys.argv:
            self._register(runner=runner)

    def _register(self, runner=None):
        if runner.model_name in self.registry:
            raise AlreadyRegistered(f"Form runner already registered. Got {runner}.")
        self.registry.update({runner.model_name: runner})

    def autodiscover(self, module_name=None, verbose=True):
        """Autodiscovers query rule classes in the form_runners.py file of
        any INSTALLED_APP.
        """
        module_name = module_name or "form_runners"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = deepcopy(site_form_runners.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f"   - registered '{module_name}' from '{app}'\n")
                except SiteFormRunnerError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_form_runners.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteFormRunnerError(str(e))
            except ImportError:
                pass


site_form_runners = SiteFormRunners()



from edc_navbar import Navbar, NavbarItem, site_navbars

device = Navbar(name="edc_device")

device.register(
    NavbarItem(
        name="device",
        label="device",
        fa_icon="fa-calculator",
        codename="edc_navbar.nav_edc_device",
        url_with_namespace="edc_device:home_url",
    )
)

site_navbars.register(device)



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin

from ...admin_site import edc_pharmacy_admin
from ...forms import ContainerForm
from ...models import Container
from ...utils import format_qty
from ..model_admin_mixin import ModelAdminMixin


@admin.register(Container, site=edc_pharmacy_admin)
class ContainerAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Containers"
    show_object_tools = True
    history_list_display = ()
    list_per_page = 20

    form = ContainerForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    [
                        "name",
                        "display_name",
                        "container_type",
                        "units",
                        "qty",
                        "qty_decimal_places",
                        "may_order_as",
                        "may_receive_as",
                        "may_repack_as",
                        "may_request_as",
                        "max_per_subject",
                        "may_dispense_as",
                    ]
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "display_name",
        "container_type",
        "formatted_units",
        "decimal_places",
        "may_order",
        "may_receive",
        "may_repack",
        "may_request",
        "max_per_subject",
        "may_dispense",
        "created",
        "modified",
    )
    list_filter = (
        "container_type",
        "units",
        "may_order_as",
        "may_receive_as",
        "may_repack_as",
        "may_request_as",
        "may_dispense_as",
        "created",
    )
    radio_fields = {"container_type": admin.VERTICAL, "units": admin.VERTICAL}  # noqa: RUF012
    search_fields = (
        "name",
        "display_name",
    )
    ordering = ("display_name",)

    @admin.display(description="Units", ordering="units")
    def formatted_units(self, obj):
        if obj.qty > 1:
            return f"{format_qty(obj.qty, obj)} {obj.units.plural_name}"
        return f"{format_qty(obj.qty, obj)} {obj.units.display_name}"

    @admin.display(description="Places", ordering="qty_decimal_places")
    def decimal_places(self, obj):
        return obj.qty_decimal_places

    @admin.display(description="order", ordering="may_order_as", boolean=True)
    def may_order(self, obj):
        return obj.may_order_as

    @admin.display(description="receive", ordering="may_receive_as", boolean=True)
    def may_receive(self, obj):
        return obj.may_receive_as

    @admin.display(description="Repack", ordering="may_repack_as", boolean=True)
    def may_repack(self, obj):
        return obj.may_repack_as

    @admin.display(description="request", ordering="may_request_as", boolean=True)
    def may_request(self, obj):
        return obj.may_request_as

    @admin.display(description="dispense", ordering="may_dispense_as", boolean=True)
    def may_dispense(self, obj):
        return obj.may_dispense_as

    def get_readonly_fields(self, request, obj=None):  # noqa: ARG002
        if obj:
            return tuple({*self.readonly_fields, "name"})
        return self.readonly_fields



from clinicedc_constants import NEW
from django.apps import apps as django_apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext as _
from edc_utils import truncate_string

from ..models import QaReportLog
from .list_filters import NoteStatusListFilter


class QaReportModelAdminMixin:
    """A mixin to link a data management report to a note (with status)
     on each report item.

    note_model_cls/template can be overridden in concrete classes.

    To use a custom Note model class, set `note_model_cls` to your
    custom model that uses the `NoteModelMixin`.
    """

    qa_report_log_enabled = True
    qa_report_list_display_insert_pos = 3
    list_per_page = 25
    note_model = "edc_qareports.note"
    note_status_list_filter = NoteStatusListFilter
    note_template = "edc_qareports/columns/notes_column.html"
    include_note_column = True

    @property
    def note_model_cls(self):
        return django_apps.get_model(self.note_model)

    def update_qa_report_log(self, request) -> None:
        QaReportLog.objects.create(
            username=request.user.username,
            site=request.site,
            report_model=self.model._meta.label_lower,
        )

    def changelist_view(self, request, extra_context=None):
        if self.qa_report_log_enabled:
            self.update_qa_report_log(request)
        return super().changelist_view(request, extra_context=extra_context)

    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        list_display = list(list_display)
        if self.include_note_column:
            list_display.insert(self.qa_report_list_display_insert_pos, "notes")
            list_display.insert(self.qa_report_list_display_insert_pos, "status")
        return tuple(list_display)

    def get_list_filter(self, request):
        list_filter = super().get_list_filter(request)
        list_filter = list(list_filter)
        if self.include_note_column:
            list_filter.insert(0, self.note_status_list_filter)
        return tuple(list_filter)

    def get_note_model_obj_or_raise(self, obj=None):
        return self.note_model_cls.objects.get(
            report_model=obj._meta.label_lower, subject_identifier=obj.subject_identifier
        )

    @admin.display(description="Status")
    def status(self, obj) -> str:
        try:
            note_model_obj = self.get_note_model_obj_or_raise(obj)
        except ObjectDoesNotExist:
            status = NEW
        else:
            status = note_model_obj.get_status_display()
        return status.title()

    @admin.display(description="Notes")
    def notes(self, obj=None) -> str:
        """Returns url to add or edit qa_report note model."""
        note_app_label, note_model_name = self.note_model_cls._meta.label_lower.split(".")
        note_url_name = f"{note_app_label}_{note_model_name}"

        report_app_label, report_model_name = self.model._meta.label_lower.split(".")
        next_url_name = "_".join([report_app_label, report_model_name, "changelist"])
        next_url_name = f"{report_app_label}_admin:{next_url_name}"

        try:
            note_model_obj = self.get_note_model_obj_or_raise(obj)
        except ObjectDoesNotExist:
            note_model_obj = None
            url = reverse(f"{note_app_label}_admin:{note_url_name}_add")
            title = _("Add if pending or cannot be resolved")
        else:
            url = reverse(
                f"{note_app_label}_admin:{note_url_name}_change",
                args=(note_model_obj.id,),
            )
            title = _("Edit if pending or cannot be resolved")

        url = (
            f"{url}?next={next_url_name},subject_identifier,q"
            f"&subject_identifier={obj.subject_identifier}"
            f"&report_model={self.model._meta.label_lower}&q={obj.subject_identifier}"
        )
        label = self.get_notes_label(note_model_obj)
        context = dict(title=title, url=url, label=label)
        return render_to_string(self.note_template, context=context)

    def get_notes_label(self, obj) -> str:
        if not obj:
            label = _("Add")
        elif not obj.note:
            label = _("Edit")
        else:
            label = truncate_string(obj.note, max_length=35)
        return label



from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_randomization/home.html"
    navbar_name = "edc_randomization"
    navbar_selected_item = "edc_randomization"



# Generated by Django 4.1.2 on 2022-11-30 01:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_sites", "0005_auto_20210423_1451"),
    ]

    operations = [
        migrations.AddField(
            model_name="siteprofile",
            name="languages",
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-23 01:13

import _socket
import django_audit_fields.fields.hostname_modification_field
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_pharmacy",
            "0019_container_extra_value_containertype_extra_value_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="box",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="box",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="box",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="dispensinghistory",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="dispensinghistory",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="dispensinghistory",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="dispensinghistory",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="dosageguideline",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="dosageguideline",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="dosageguideline",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="dosageguideline",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="formulation",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="formulation",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="formulation",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="formulation",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="genericcontainer",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="genericcontainer",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="genericcontainer",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="genericcontainer",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldispensinghistory",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicaldispensinghistory",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicaldispensinghistory",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldispensinghistory",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldosageguideline",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicaldosageguideline",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicaldosageguideline",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldosageguideline",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalformulation",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalformulation",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalformulation",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalformulation",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicallabels",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicallabels",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicallabels",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicallabels",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmedication",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalmedication",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalmedication",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmedication",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmedicationlot",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalmedicationlot",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalmedicationlot",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmedicationlot",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalproduct",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalproduct",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalproduct",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalproduct",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreturnhistory",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalreturnhistory",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalreturnhistory",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreturnhistory",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrx",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalrx",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalrx",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrx",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockcreatelabels",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalstockcreatelabels",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalstockcreatelabels",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockcreatelabels",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockreceiving",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalstockreceiving",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalstockreceiving",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockreceiving",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="labels",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="labels",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="labels",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="labels",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="location",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="location",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="location",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="location",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="medication",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="medication",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="medication",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="medication",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="medicationlot",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="medicationlot",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="medicationlot",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="medicationlot",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="order",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="order",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="order",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="pillbottle",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="pillbottle",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="pillbottle",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="pillbottle",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="product",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="product",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="returnhistory",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="returnhistory",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="returnhistory",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="returnhistory",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="room",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="room",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="room",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="room",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="rx",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="rx",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="rx",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="rx",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="rxrefill",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="shelf",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="shelf",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="shelf",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="shelf",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="stock",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="stock",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="stockcreatelabels",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="stockcreatelabels",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="stockcreatelabels",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="stockcreatelabels",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="stockreceiving",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="stockreceiving",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="stockreceiving",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="stockreceiving",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectpillbottle",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="subjectpillbottle",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="subjectpillbottle",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="subjectpillbottle",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
    ]



class AliquotIdentifierLengthError(Exception):
    pass


class AliquotIdentifierCountError(Exception):
    pass


class AliquotIdentifier:
    count_padding: int = 2
    identifier_length: int = 18
    primary_aliquot_segment: str = "0000"
    template: str = "{identifier_prefix}{parent_segment}{numeric_code}{count}"

    def __init__(
        self,
        identifier_prefix: str | None = None,
        parent_segment: str | None = None,
        numeric_code: str | None = None,
        count: int | None = None,
    ):
        """
        A class to generate aliquot identifiers:

        Keyword args:
            * identifier_prefix: a prefix as instance of Prefix
            * numeric_code: aliquot type numeric code (2 digits segment)
            * count: sequence in aliquoting history relative to primary. (01 for primary)
        """
        self.count: int | None = count
        self.identifier_prefix: str = identifier_prefix or ""
        self.is_primary: bool = True
        self.numeric_code: str = numeric_code or ""
        self.parent_segment: str = parent_segment or ""
        if not self.identifier_length:
            raise AliquotIdentifierLengthError(
                f"Invalid length. Got {self.identifier_length}."
            )
        if self.parent_segment:
            self.is_primary = False
            if not self.count or self.count <= 1:
                raise AliquotIdentifierCountError(
                    "Unknown aliquot number/count. Expected a number "
                    f"greater than 1. Got {self.count}."
                )
        else:
            self.parent_segment = self.primary_aliquot_segment
            self.count = 1
        self.identifier: str = self.template.format(**self.options)
        if len(self.identifier) != self.identifier_length:
            raise AliquotIdentifierLengthError(
                f"Invalid length. Expected {self.identifier_length}. "
                f"Got len({self.identifier})=={len(self.identifier)}."
            )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.identifier_length})"

    def __str__(self):
        return self.identifier

    @property
    def options(self) -> dict:
        return dict(
            identifier_prefix=self.identifier_prefix,
            parent_segment=self.parent_segment,
            numeric_code=self.numeric_code,
            count=str(self.count).zfill(self.count_padding or 0),
        )



from edc_model_admin.admin_site import EdcAdminSite as BaseEdcAdminSite

from .apps import AppConfig


class EdcAdminSite(BaseEdcAdminSite):
    index_template = "edc_adverse_event/admin/index.html"
    app_index_template = "edc_adverse_event/admin/app_index.html"


edc_adverse_event_admin = EdcAdminSite(
    name="edc_adverse_event_admin", app_label=AppConfig.name
)



from __future__ import annotations

from typing import TYPE_CHECKING

from reportlab.graphics.barcode.widgets import BarcodeCode128
from reportlab.graphics.shapes import Drawing, Group, String
from reportlab.pdfbase.pdfmetrics import stringWidth

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from ..utils import format_qty
from .draw_label_watermark import draw_label_watermark

if TYPE_CHECKING:
    from ..models import Stock


def draw_vertical_barcode_only_code128(
    label: Drawing,
    width: int | float,
    height: int | float,
    obj: Stock,
) -> Drawing:

    draw_label_watermark(label, width, height)

    br = BarcodeCode128(humanReadable=True, barHeight=30, barWidth=0.7, gap=1.7)
    br.value = obj.code
    br.x = 0
    br.y = height - 68
    group = Group()
    group.add(br)
    group.translate(width - 170, height - 108)
    group.rotate(90)
    label.add(group)

    protocol_name = String(0, 0, str(ResearchProtocolConfig().protocol_name))
    qty_text = f"{format_qty(obj.container.qty, obj.container)} tabs"

    text_group = Group()
    text_group.add(protocol_name)

    text_width = stringWidth(qty_text, "Helvetica", 10)
    text_string = String(height - text_width - 10, 0, qty_text)
    text_group.add(text_string)

    text_group.translate(10, height - 100)
    text_group.rotate(90)
    label.add(text_group)

    return label



# group names
QA_REPORTS = "QA_REPORTS"
QA_REPORTS_AUDIT = "QA_REPORTS_AUDIT"

# role names
QA_REPORTS_ROLE = "qa_reports_role"
QA_REPORTS_SUPER_ROLE = "qa_reports_super_role"
QA_REPORTS_AUDIT_ROLE = "qa_reports_audit_role"

qa_reports_codenames = [
    "edc_qareports.add_note",
    "edc_qareports.change_note",
    "edc_qareports.delete_note",
    "edc_qareports.view_note",
    "edc_qareports.view_historicalnote",
    "edc_qareports.viewallsites_note",
    "edc_qareports.view_qareportlog",
    "edc_qareports.viewallsites_qareportlog",
    "edc_qareports.view_qareportlogsummary",
    "edc_qareports.viewallsites_qareportlogsummary",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.contrib import admin, messages
from django.db.models import Count, QuerySet
from django.http import FileResponse
from django.utils import timezone
from django.utils.translation import gettext as _
from pylabels import Sheet, Specification

from .site_label_configs import LabelConfig, site_label_configs

if TYPE_CHECKING:
    from .models import LabelConfiguration


@admin.action(description="Test print sheet of labels")
def print_test_label_sheet_action(modeladmin, request, queryset: QuerySet[LabelConfiguration]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            _("Select one and only one existing label specification"),
        )
    else:
        obj = queryset.first()
        label_config: LabelConfig = site_label_configs.get(obj.name)
        specs = Specification(**obj.label_specification.as_dict)
        sheet = Sheet(
            specs, label_config.drawing_callable, border=obj.label_specification.border
        )

        try:
            label_cls = django_apps.get_model(label_config.label_cls)
        except (LookupError, AttributeError):
            label_cls = label_config.label_cls
        data = label_config.test_data_func()
        sheet.add_labels(
            [
                label_cls(**data)
                for i in range(
                    0, obj.label_specification.rows * obj.label_specification.columns
                )
            ]
        )
        buffer = sheet.save_to_buffer()
        return FileResponse(buffer, as_attachment=True, filename=f"test_print_{obj.name}.pdf")
    return None


@admin.action(description="Print labels")
def print_label_sheet(modeladmin, request, queryset):
    if (
        queryset.model.objects.values("label_configuration__name")
        .filter(id__in=[obj.id for obj in queryset])
        .annotate(name=Count("label_configuration__name"))
        .count()
        > 1
    ):
        messages.add_message(
            request,
            messages.ERROR,
            _("Select items that have the same label configuration"),
        )
    else:
        label_data = [obj for obj in queryset]
        config: LabelConfiguration = queryset.first().label_configuration
        drawing_callable = site_label_configs.get(config.name).drawing_callable
        specs = Specification(**config.label_specification.as_dict)
        sheet = Sheet(specs, drawing_callable, border=config.label_specification.border)
        sheet.add_labels(label_data)
        buffer = sheet.save_to_buffer()
        now = timezone.now()
        return FileResponse(
            buffer,
            as_attachment=True,
            filename=f"{config.name}_{now.strftime('%Y-%m-%d %H:%M')}.pdf",
        )
    return None



from django.db import models


class VisitScheduleFieldsModelMixin(models.Model):
    """A model mixin that adds fields required to work with the visit
    schedule methods on the VisitScheduleMethodsModelMixin.

    Note: visit_code is not included.
    """

    visit_schedule_name = models.CharField(
        max_length=25,
        editable=False,
        help_text='the name of the visit schedule used to find the "schedule"',
    )

    schedule_name = models.CharField(max_length=25, editable=False)

    class Meta:
        abstract = True



# Generated by Django 5.1.2 on 2024-11-16 17:32

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0030_remove_historicalrepackrequest_processed_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalrepackrequest",
            name="task_id",
            field=models.UUIDField(null=True),
        ),
        migrations.AddField(
            model_name="repackrequest",
            name="task_id",
            field=models.UUIDField(null=True),
        ),
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="processed_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Containers processed",
            ),
        ),
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="requested_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Containers requested",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="processed_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Containers processed",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="requested_qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Containers requested",
            ),
        ),
    ]



from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin

from ..dashboard_templates import dashboard_templates


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = dashboard_templates.get("home_template")
    navbar_name = "specimens"
    navbar_selected_item = "specimens"



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_pharmacy_admin
from ..models import ScanDuplicates


@admin.register(ScanDuplicates, site=edc_pharmacy_admin)
class ScanDuplicatesAdmin(SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Scan duplicates"
    change_form_title = "Pharmacy: Scan duplicates"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    fieldsets = (
        (
            None,
            {"fields": ("identifier",)},
        ),
        audit_fieldset_tuple,
    )

    list_display = ("identifier",)

    search_fields = ("id", "identifier")



from django.conf import settings

from .base_form_validator import BaseFormValidator


class CurrentSiteValidator(BaseFormValidator):
    def __init__(self, current_site=None, **kwargs):
        self.current_site = current_site or settings.SITE_ID
        super().__init__(**kwargs)



import uuid

from django.db import models
from simple_history.models import HistoricalRecords as SimpleHistoricalRecords


class SerializableModelManager(models.Manager):
    def get_by_natural_key(self, history_id):
        return self.get(history_id=history_id)


class SerializableModel(models.Model):
    objects = SerializableModelManager()

    def natural_key(self) -> tuple:
        return (self.history_id,)

    def related_visit_model_attr(self):
        return self.history_object.related_visit_model_attr()

    @property
    def related_visit(self):
        return getattr(self, self.related_visit_model_attr())

    class Meta:
        abstract = True


class SerializableCrfModel(models.Model):
    objects = SerializableModelManager()

    @property
    def related_visit(self):
        return NotImplemented(self)

    def natural_key(self) -> tuple:
        return (self.history_id,)

    class Meta:
        abstract = True


class HistoricalRecords(SimpleHistoricalRecords):
    """HistoricalRecords that forces a UUID primary key,
    has a natural key method available for serialization,
    and respects \'using\'.
    """

    model_cls = SerializableModel

    def __init__(self, **kwargs):
        """Defaults use of UUIDField instead of AutoField and
        serializable base.
        """
        kwargs.update(bases=(self.model_cls,))
        kwargs.update(history_id_field=models.UUIDField(default=uuid.uuid4))
        kwargs.update(use_base_model_db=True)
        super().__init__(**kwargs)

    def contribute_to_class(self, cls, name):
        if getattr(cls, "related_visit_model_attr", None):
            self.model_cls = SerializableCrfModel
        return super().contribute_to_class(cls, name)



from edc_navbar import Navbar, NavbarItem, site_navbars

protocol = Navbar(name="edc_protocol")

protocol.register(
    NavbarItem(
        name="protocol",
        title="Protocol",
        label="protocol",
        codename="edc_navbar.nav_edc_protocol",
        url_with_namespace="edc_protocol:home_url",
    )
)

site_navbars.register(protocol)



import random

random.seed(3897538743783)


def get_random_code(model_cls, length: int, tries: int | None = None) -> str:
    random_code = 0
    x = 0
    tries = tries or 1000
    while x < tries:
        random_code = str(
            "".join(
                [
                    random.choice("ABCDEFGHJKMNPQRTUVWXYZ2346789")  # nosec B311
                    for _ in range(0, length)
                ]
            )
        )
        if not model_cls.objects.filter(code=random_code).exists():
            break
        x += 1
        if x == tries:
            raise StopIteration()
    return random_code


__all__ = ["get_random_code"]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.utils.translation import gettext as _

from edc_consent.consent_definition import ConsentDefinition
from edc_consent.exceptions import (
    ConsentDefinitionDoesNotExist,
    ConsentDefinitionNotConfiguredForUpdate,
    NotConsentedError,
    SiteConsentError,
)
from edc_consent.site_consents import site_consents
from edc_form_validators import INVALID_ERROR
from edc_sites.site import sites as site_sites

if TYPE_CHECKING:
    from edc_consent.stubs import ConsentLikeModel


class ConsentDefinitionFormValidatorMixin:
    @property
    def subject_consent(self):
        cdef = self.get_consent_definition(
            self.report_datetime, self.report_datetime_field_attr
        )
        return cdef.model_cls.objects.get(subject_identifier=self.subject_identifier)

    def get_consent_datetime_or_raise(
        self,
        reference_datetime: datetime | None = None,
        reference_datetime_field: str | None = None,
        error_code: str | None = None,
    ) -> datetime:
        """Returns the consent_datetime of this subject"""
        consent_obj = self.get_consent_or_raise(
            reference_datetime=reference_datetime,
            reference_datetime_field=reference_datetime_field,
            error_code=error_code,
        )
        return consent_obj.consent_datetime

    def get_consent_or_raise(
        self,
        reference_datetime: datetime | None = None,
        reference_datetime_field: str | None = None,
        error_code: str | None = None,
    ) -> ConsentLikeModel:
        """Returns the consent_datetime of this subject.

        Wraps func `consent_datetime_or_raise` to re-raise exceptions
        as ValidationError.
        """

        error_code = error_code or INVALID_ERROR
        try:
            consent_obj = site_consents.get_consent_or_raise(
                subject_identifier=self.subject_identifier,
                report_datetime=reference_datetime or self.report_datetime,
                site_id=getattr(self.site, "id", None),
            )
        except (NotConsentedError, ConsentDefinitionNotConfiguredForUpdate) as e:
            self.raise_validation_error(
                {reference_datetime_field or self.report_datetime_field_attr: str(e)},
                error_code,
                exc=e,
            )
        return consent_obj

    def get_consent_definition(
        self,
        report_datetime: datetime,
        fldname: str = None,
        error_code: str | None = None,
    ) -> ConsentDefinition:
        error_code = error_code or INVALID_ERROR
        # get the consent definition (must be from this schedule)
        schedule = getattr(self, "related_visit", self.instance).schedule

        site = getattr(self, "related_visit", self.instance).site
        try:
            consent_definition = schedule.get_consent_definition(
                site=site_sites.get(site.id),
                report_datetime=report_datetime,
            )
        except ConsentDefinitionDoesNotExist as e:
            self.raise_validation_error({fldname: str(e)}, error_code, exc=e)
        except SiteConsentError as e:
            possible_consents = "', '".join(
                [cdef.display_name for cdef in site_consents.consent_definitions]
            )
            msg = _(
                "Date does not fall within a valid consent period. "
                "Possible consents are '%(possible_consents)s'. "
            ) % {"possible_consents": possible_consents}

            self.raise_validation_error({fldname: msg}, error_code, exc=e)
        return consent_definition



import pandas as pd
from clinicedc_constants import FEMALE, MALE

from ..constants import MEDIAN_RANGE, N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class AgeTable(Table):

    col_a = FEMALE
    col_b = MALE
    column_name = "gender"

    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(colname="age_in_years", main_df=main_df, title="Age (years)")

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp[self.column_name].notna()),
            columns={
                self.col_a: (N_ONLY, 2),
                self.col_b: (N_ONLY, 2),
                "All": (N_ONLY, 2),
            },
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            self.col_a: (N_WITH_COL_PROP, 2),
            self.col_b: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }
        bin1 = (df_tmp[self.colname] >= 18) & (df_tmp[self.colname] < 35)
        bin2 = (df_tmp[self.colname] >= 35) & (df_tmp[self.colname] < 50)
        bin3 = (df_tmp[self.colname] >= 50) & (df_tmp[self.colname] < 65)
        bin4 = df_tmp[self.colname] >= 65
        row_defs.add(RowDefinition(label="18-34", condition=bin1, columns=columns, drop=False))
        row_defs.add(RowDefinition(label="35-49", condition=bin2, columns=columns, drop=False))
        row_defs.add(RowDefinition(label="50-64", condition=bin3, columns=columns, drop=False))
        row_defs.add(
            RowDefinition(label="65 and older", condition=bin4, columns=columns, drop=False)
        )
        if len(df_tmp[df_tmp[self.colname].isna()]) > 0:
            row_defs.add(
                RowDefinition(
                    colname="age_in_years",
                    label="not recorded",
                    condition=(df_tmp[self.colname].isna()),
                    columns=columns,
                    drop=False,
                )
            )
        columns = {
            self.col_a: (MEDIAN_RANGE, 2),
            self.col_b: (MEDIAN_RANGE, 2),
            "All": (MEDIAN_RANGE, 2),
        }
        row_defs.add(
            RowDefinition(
                colname="age_in_years",
                label="Median (range)",
                condition=(self.main_df[self.colname].notna()),
                columns=columns,
            )
        )
        return row_defs



from django import forms

from ..form_validators import AeTmgFormValidator
from . import AeModelFormMixin


class AeTmgModelFormMixin(AeModelFormMixin):
    form_validator_cls = AeTmgFormValidator

    class Meta:
        labels = {  # noqa: RUF012
            "ae_description": "Original AE Description",
            "ae_classification": "AE Classification",
            "ae_classification_other": "AE Classification (if `other` above)",
        }
        help_text = {  # noqa: RUF012
            "subject_identifier": "(read-only)",
            "action_identifier": "(read-only)",
            "ae_description": "(read-only)",
            "ae_classification": "(read-only)",
            "ae_classification_other": "(read-only)",
        }
        widgets = {  # noqa: RUF012
            "subject_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "action_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "ae_description": forms.Textarea(attrs={"readonly": "readonly", "cols": "79"}),
            "ae_classification": forms.TextInput(attrs={"readonly": "readonly"}),
            "ae_classification_other": forms.TextInput(attrs={"readonly": "readonly"}),
        }



# Generated by Django 6.0 on 2026-01-19 17:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0103_auto_20260119_1945"),
    ]

    operations = [
        migrations.AddField(
            model_name="allocation",
            name="subject_identifier",
            field=models.CharField(
                blank=True, default="", editable=False, max_length=50
            ),
        ),
        migrations.AddField(
            model_name="historicalallocation",
            name="subject_identifier",
            field=models.CharField(
                blank=True, default="", editable=False, max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="subject_identifier",
            field=models.CharField(
                blank=True, default="", editable=False, max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="subject_identifier",
            field=models.CharField(
                blank=True, default="", editable=False, max_length=50
            ),
        ),
    ]



from clinicedc_constants import OTHER

list_data = {
    "edc_refusal.refusalreasons": [
        ("unwilling_to_say", "I am unwilling to say"),
        ("dont_have_time", "I don't have time"),
        ("stigma", "I am worried about stigma"),
        ("must_consult_spouse", "I need to consult my spouse"),
        ("dont_want_medication", "I don't want to take any more medication"),
        ("dont_want_to_join", "I don't want to take part"),
        ("no_clinic_time", "I don't want to spend a lot of time at the clinic"),
        ("need_to_think_about_it", "I haven't had a chance to think about it"),
        ("moving", "I am moving to another area"),
        (OTHER, "Other, please specify"),
    ],
}



from __future__ import annotations

import re
from typing import Any

from django.db.models import Q

from edc_dashboard.view_mixins import EdcViewMixin
from edc_listboard.view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from edc_listboard.views import ListboardView as BaseListboardView
from edc_navbar import NavbarViewMixin

from ...utils import get_adverse_event_app_label


class DeathListboardView(
    NavbarViewMixin,
    EdcViewMixin,
    ListboardFilterViewMixin,
    SearchFormViewMixin,
    BaseListboardView,
):
    subject_dashboard_url_name = "subject_dashboard_url"  # see url_names
    listboard_back_url = "tmg_home_url"

    listboard_template = "tmg_death_listboard_template"
    listboard_url = "tmg_death_listboard_url"  # url_name
    listboard_panel_style = "warning"
    listboard_model = f"{get_adverse_event_app_label()}.deathreport"
    listboard_model_manager_name = "objects"
    listboard_panel_title = "TMG: Death Reports"
    listboard_view_permission_codename = "edc_adverse_event.view_tmg_listboard"
    navbar_selected_item = "tmg_home"
    ordering = "-created"
    paginate_by = 25
    search_form_url = "tmg_death_listboard_url"  # url_name
    search_fields = (
        "subject_identifier",
        "action_identifier",
        "parent_action_item__action_identifier",
        "related_action_item__action_identifier",
        "user_created",
        "user_modified",
    )

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        # kwargs.update(
        #     {"subject_dashboard_url": url_names.get(self.subject_dashboard_url_name)}
        # )
        if self.kwargs.get("subject_identifier"):
            kwargs.update({"q": self.kwargs.get("subject_identifier")})
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        if self.search_term and re.match("^[A-Z]+$", self.search_term):
            q_object |= Q(first_name__exact=self.search_term)
        if kwargs.get("subject_identifier"):
            options.update({"subject_identifier": kwargs.get("subject_identifier")})
        return q_object, options



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand
from django.core.management.color import color_style
from tqdm import tqdm

from edc_appointment.skip_appointments import (
    SkipAppointments,
    SkipAppointmentsValueError,
)
from edc_appointment.utils import get_allow_skipped_appt_using
from edc_registration.models import RegisteredSubject

style = color_style()


class Command(BaseCommand):
    help = "Update skipped appointments"

    def handle(self, *args, **options) -> None:  # noqa: ARG002
        errors: dict[str, list[str]] = {}
        for model in get_allow_skipped_appt_using():
            crf_model_cls = django_apps.get_model(model)
            qs = RegisteredSubject.objects.all().order_by("subject_identifier")
            total = qs.count()
            for subject_identifier in tqdm(qs, total=total):
                for subject_visit in (
                    crf_model_cls.related_visit_model_cls()
                    .objects.filter(
                        subject_identifier=subject_identifier, visit_code_sequence=0
                    )
                    .order_by("report_datetime")
                ):
                    try:
                        crf_obj = crf_model_cls.objects.get(subject_visit=subject_visit)
                    except ObjectDoesNotExist:
                        pass
                    else:
                        try:
                            SkipAppointments(crf_obj).update()
                        except SkipAppointmentsValueError as e:
                            msg = (
                                f"{e}. See {subject_visit.subject_identifier}"
                                f"@{subject_visit.visit_code}."
                            )
                            try:
                                errors[subject_visit.subject_identifier].append(msg)
                            except KeyError:
                                errors.update({subject_visit.subject_identifier: [msg]})
                            print(msg)  # noqa: T201
        print("\nERRORS\n")  # noqa: T201
        for k, v in errors.items():
            print(f"{k} ---------------")  # noqa: T201
            for msg in v:
                print(msg)  # noqa: T201
        print("\n\nDone")  # noqa: T201



from edc_model_admin.admin_site import EdcAdminSite

from .apps import AppConfig

edc_data_manager_admin = EdcAdminSite(
    name="edc_data_manager_admin", app_label=AppConfig.name, keep_delete_action=True
)



from decimal import Decimal

from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import Sum
from django.utils import timezone

from edc_model.models import BaseUuidModel, HistoricalRecords

from .stock import Stock


class StockAdjustmentManager(models.Manager):
    pass


class StockAdjustment(BaseUuidModel):
    stock = models.ForeignKey(
        Stock,
        # related_name="source_stock",
        on_delete=models.PROTECT,
        null=True,
    )

    adjustment_datetime = models.DateTimeField(default=timezone.now)

    unit_qty_in_old = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0)],
    )

    unit_qty_in_new = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0)],
    )

    reason = models.TextField(default="")

    objects = StockAdjustmentManager()

    history = HistoricalRecords()

    def __str__(self):
        return (
            f"Stock Adjustment for {self.stock.code}. "
            f"{self.unit_qty_in_old}->{self.unit_qty_in_new}."
        )

    def save(self, *args, **kwargs):
        self.unit_qty_in_new = (
            Stock.objects.filter(from_stock=self.stock)
            .aggregate(unit_qty_in=Sum("unit_qty_in"))
            .get("unit_qty_in")
        )
        super().save(*args, **kwargs)

    class Meta:
        verbose_name_plural = "Stock Adjustments"
        verbose_name = "Stock Adjustment"



from django.db import models
from django.db.models import PROTECT


class MedicationOrderModelMixin(models.Model):
    stock = models.ForeignKey(
        "edc_pharmacy.stock",
        null=True,
        blank=False,
        on_delete=PROTECT,
    )

    qty = models.DecimalField(null=True, blank=False, decimal_places=2, max_digits=10)

    packed = models.BooleanField(default=False)
    packed_datetime = models.DateTimeField(null=True, blank=True)

    shipped = models.BooleanField(default=False)
    shipped_datetime = models.DateTimeField(null=True, blank=True)

    received_at_site = models.BooleanField(default=False)
    received_at_site_datetime = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True



# Generated by Django 3.2.13 on 2022-09-22 19:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0030_edcpermissions"),
    ]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="reference_model",
            field=models.CharField(
                help_text="model name in label_lower format", max_length=50, null=True
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="reference_model",
            field=models.CharField(
                help_text="model name in label_lower format", max_length=50, null=True
            ),
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils import timezone
from django_crypto_fields.fields import EncryptedTextField

from edc_action_item.models import ActionModelMixin
from edc_constants.choices import YES_NO, YES_NO_UNSURE
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_model import models as edc_models
from edc_sites.model_mixins import SiteModelMixin
from edc_utils.text import convert_php_dateformat

from .choices import TRANSFER_INITIATORS
from .constants import SUBJECT_TRANSFER_ACTION


class SubjectTransferModelMixin(
    SiteModelMixin,
    ActionModelMixin,
    UniqueSubjectIdentifierFieldMixin,
    models.Model,
):
    action_name = SUBJECT_TRANSFER_ACTION

    report_datetime = models.DateTimeField(
        verbose_name="Report Date and Time", default=timezone.now
    )

    transfer_date = models.DateField(verbose_name="Transfer date", default=timezone.now)

    initiated_by = models.CharField(
        verbose_name="Who initiated the transfer request",
        max_length=25,
        choices=TRANSFER_INITIATORS,
    )

    initiated_by_other = edc_models.OtherCharField()

    transfer_reason = models.ManyToManyField(
        "edc_transfer.TransferReasons",
        verbose_name="Reason for transfer",
    )

    transfer_reason_other = edc_models.OtherCharField()

    may_return = models.CharField(
        verbose_name=(
            "Is the participant likely to transfer back before "
            "the end of their stay in the trial?"
        ),
        max_length=15,
        choices=YES_NO_UNSURE,
    )

    may_contact = models.CharField(
        verbose_name="Is the participant willing to be contacted at the end of the study?",
        max_length=15,
        choices=YES_NO,
    )

    comment = EncryptedTextField(verbose_name="Additional Comments")

    def __str__(self):
        transfer_date = "???"
        if self.transfer_date:
            transfer_date = self.transfer_date.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
        return f"{self.subject_identifier} on {transfer_date}."

    def natural_key(self):
        return (self.subject_identifier,)

    class Meta(SiteModelMixin.Meta, ActionModelMixin.Meta):
        abstract = True
        verbose_name = "Subject Transfer"
        verbose_name_plural = "Subject Transfers"
        indexes = (
            *ActionModelMixin.Meta.indexes,
            models.Index(
                fields=[
                    "subject_identifier",
                    "action_identifier",
                    "report_datetime",
                    "site",
                ]
            ),
        )



from django.apps import apps as django_apps
from django.contrib import admin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import ModelAdminInstitutionMixin, TemplatesModelAdminMixin
from edc_sites.admin import SiteModelAdminMixin

from ..admin_site import edc_qareports_admin
from ..models import QaReportLogSummary


@admin.register(QaReportLogSummary, site=edc_qareports_admin)
class QaReportLogSummaryAdmin(
    SiteModelAdminMixin,
    ModelAdminRevisionMixin,
    ModelAdminInstitutionMixin,
    TemplatesModelAdminMixin,
    admin.ModelAdmin,
):
    ordering = ("-last_accessed",)

    list_display = (
        "username",
        "report",
        "hits",
        "first",
        "last",
        "report_model",
    )

    list_filter = ("last_accessed", "report_model", "username")

    search_fields = ("report_model", "username")

    readonly_fields = (
        "report_model",
        "username",
        "site",
        "first_accessed",
        "last_accessed",
        "access_count",
    )

    @admin.display(description="Report", ordering="report_model")
    def report(self, obj=None):
        try:
            model_cls = django_apps.get_model(obj.report_model)
        except LookupError:
            return obj.report_model
        return model_cls._meta.verbose_name

    @admin.display(description="Hits", ordering="access_count")
    def hits(self, obj=None):
        return obj.access_count

    @admin.display(description="First", ordering="first_accessed")
    def first(self, obj=None):
        return obj.first_accessed.strftime("%Y-%m-%d")

    @admin.display(description="Last", ordering="last_accessed")
    def last(self, obj=None):
        return obj.last_accessed.strftime("%Y-%m-%d")



from dataclasses import dataclass, field
from datetime import datetime

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .address import Address


class EdcProtocolError(Exception):
    pass


@dataclass
class ResearchProtocolConfig:
    """Encapsulates settings attributes.

    * EDC_PROTOCOL: 6 digit alphanumeric
    * EDC_PROTOCOL_INSTITUTION_NAME
    * EDC_PROTOCOL_NUMBER: Used for identifiers NNN
    * EDC_PROTOCOL_PROJECT_NAME: Short name
        e.g. Mashi, Tshepo, Ambition, BCPP, META, INTE, etc
    * EDC_PROTOCOL_STUDY_CLOSE_DATETIME
    * EDC_PROTOCOL_STUDY_OPEN_DATETIME
    * EDC_PROTOCOL_TITLE: Long name
    * EMAIL_CONTACTS
    """

    error_msg1: str = field(
        default=(
            "Unable to set `%(attr)s`. "
            "settings.%(settings_attr)s not found. "
            "Expected something like: `%(settings_attr)s = "
            'datetime(2013, 10, 15, tzinfo=ZoneInfo("Africa/Gaborone"))`. '
            "See edc_protocol."
        ),
        init=False,
        repr=False,
    )
    error_msg2: str = field(
        default=(
            "Unable to set `%(attr)s`. "
            "Settings.%(settings_attr)s cannot be None. "
            "Expected something like: `%(settings_attr)s = "
            'datetime(2013, 10, 15, tzinfo=ZoneInfo("Africa/Gaborone"))`. '
            "See edc_protocol."
        ),
        init=False,
        repr=False,
    )

    @property
    def trial_group(self) -> str:
        return getattr(settings, "EDC_PROTOCOL_TRIAL_GROUP", "-")

    @property
    def protocol(self) -> str:
        return getattr(settings, "EDC_PROTOCOL", "AAA000")

    @property
    def protocol_number(self) -> str:
        """3 digits, used for identifiers, required for live systems"""
        protocol_number = getattr(settings, "EDC_PROTOCOL_NUMBER", "000")
        if not settings.DEBUG and protocol_number == "000":
            raise EdcProtocolError(
                "Settings attribute `EDC_PROTOCOL_NUMBER` not defined or "
                "set to '000' while DEBUG=False."
            )

        return protocol_number

    @property
    def protocol_title(self) -> str:
        return getattr(
            settings, "EDC_PROTOCOL_TITLE", "Protocol Title (set EDC_PROTOCOL_TITLE)"
        )

    @property
    def email_contacts(self) -> str:
        return getattr(settings, "EMAIL_CONTACTS", {})

    @property
    def institution(self) -> str:
        return getattr(
            settings,
            "EDC_PROTOCOL_INSTITUTION_NAME",
            "Institution (set EDC_PROTOCOL_INSTITUTION_NAME)",
        )

    @property
    def project_name(self) -> str:
        return getattr(
            settings,
            "EDC_PROTOCOL_PROJECT_NAME",
            "Project Name (set EDC_PROTOCOL_PROJECT_NAME)",
        )

    @property
    def protocol_name(self) -> str:
        return self.project_name

    @property
    def protocol_lower_name(self) -> str:
        return "_".join(self.protocol_name.lower().split(" "))

    @property
    def disclaimer(self) -> str:
        return _("For research purposes only")

    @property
    def copyright(self) -> str:
        return f"2010-{timezone.now().year}"

    @property
    def license(self) -> str:
        return "GNU GENERAL PUBLIC LICENSE Version 3"

    @property
    def default_url_name(self) -> str:
        return "home_url"

    @property
    def physical_address(self) -> Address:
        return Address()

    @property
    def postal_address(self) -> Address:
        return Address()

    @property
    def subject_identifier_pattern(self) -> str:
        return getattr(
            settings,
            "EDC_PROTOCOL_SUBJECT_IDENTIFIER_PATTERN",
            r"{protocol_number}\-[0-9\-]+".format(
                **dict(protocol_number=self.protocol_number)
            ),
        )

    @property
    def screening_identifier_pattern(self) -> str:
        return getattr(settings, "EDC_PROTOCOL_SCREENING_IDENTIFIER_PATTERN", r"[A-Z0-9]{8}")

    @property
    def study_open_datetime(self) -> datetime:
        try:
            study_open_datetime = settings.EDC_PROTOCOL_STUDY_OPEN_DATETIME
        except AttributeError as e:
            raise ImproperlyConfigured(
                self.error_msg1
                % {
                    "attr": "study_open_datetime",
                    "settings_attr": "EDC_PROTOCOL_STUDY_OPEN_DATETIME",
                }
            ) from e
        if not study_open_datetime:
            raise ImproperlyConfigured(
                self.error_msg2
                % {
                    "attr": "study_open_datetime",
                    "settings_attr": "EDC_PROTOCOL_STUDY_OPEN_DATETIME",
                }
            )
        return study_open_datetime

    @property
    def study_close_datetime(self) -> datetime:
        try:
            study_close_datetime = settings.EDC_PROTOCOL_STUDY_CLOSE_DATETIME
        except AttributeError as e:
            raise ImproperlyConfigured(
                self.error_msg1
                % {
                    "attr": "study_close_datetime",
                    "settings_attr": "EDC_PROTOCOL_STUDY_CLOSE_DATETIME",
                }
            ) from e
        if not study_close_datetime:
            raise ImproperlyConfigured(
                self.error_msg2
                % {
                    "attr": "study_close_datetime",
                    "settings_attr": "EDC_PROTOCOL_STUDY_CLOSE_DATETIME",
                }
            )
        return study_close_datetime



from __future__ import annotations

from typing import TYPE_CHECKING

from reportlab.graphics.barcode.widgets import BarcodeCode128
from reportlab.graphics.shapes import Drawing, String

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from ..utils import format_qty
from .draw_label_watermark import draw_label_watermark

if TYPE_CHECKING:
    from ..models import Stock


def draw_bulk_stock_label_code128(
    label: Drawing,
    width: int | float,
    height: int | float,
    obj: Stock,
) -> Drawing:
    """Callable to draw a single study medication label given a model
    instance `obj`
    """

    draw_label_watermark(label, width, height, fontSize=18)

    text = str(ResearchProtocolConfig().protocol_name)
    qty = format_qty(obj.container.qty, obj.container)
    br = BarcodeCode128(humanReadable=True, barHeight=30, barWidth=0.7, gap=1.7)
    br.value = obj.code
    br.x = width - 100
    br.y = height - 40
    label.add(br)

    label.add(String(15, height - 20, text, fontSize=10))
    label.add(String(15, height - 42, f"Lot: {obj.lot.lot_no}", fontSize=10))
    label.add(
        String(
            15,
            height - 54,
            f"Receive#: {obj.receive_item.receive.receive_identifier}",
            fontSize=10,
        )
    )
    label.add(String(15, height - 66, f"Expires: {obj.lot.expiration_date}", fontSize=10))

    formulation = obj.product.formulation
    text = (
        f"{formulation.medication} {obj.product.assignment.name.upper()} "
        f"{int(formulation.strength)}{formulation.get_units_display()}"
    )
    label.add(String(15, height - 84, text, fontSize=10))
    label.add(String(15, height - 96, f"{qty} tabs", fontSize=10))
    return label



from django import forms
from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ...admin_site import edc_pharmacy_admin
from ...models import Assignment
from ..model_admin_mixin import ModelAdminMixin


class AssignmentForm(forms.ModelForm):
    class Meta:
        model = Assignment
        fields = "__all__"


@admin.register(Assignment, site=edc_pharmacy_admin)
class AssignmentAdmin(ModelAdminMixin, admin.ModelAdmin):
    show_object_tools = True
    show_cancel = True

    form = AssignmentForm

    fieldsets = (
        (
            None,
            {"fields": ["name", "display_name"]},
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "name",
        "display_name",
        "created",
        "modified",
    )

    search_fields = ("name", "display_name")

    def get_readonly_fields(self, request, obj=None):  # noqa: ARG002
        if obj:
            return tuple({*self.readonly_fields, "name"})
        return self.readonly_fields



def insert_into_domain(domain: str, subdomain: str):
    as_list = domain.split(".")
    if subdomain not in as_list:
        as_list.insert(1, subdomain)  # after the site name
    return ".".join(as_list)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.views.generic.base import TemplateView
from edc_dashboard.view_mixins import EdcViewMixin

from ..constants import CSV
from ..exportable_models_for_user import ExportableModelsForUser
from ..files_emailer import FilesEmailerError
from ..model_options import ModelOptions
from ..models_to_file import ModelsToFile, ModelsToFileNothingExportedError
from ..utils import email_files_to_user, update_data_request_history

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class ExportModelsViewError(Exception):
    pass


class ExportSelectedModelsView(EdcViewMixin, TemplateView):
    post_action_url = "edc_export:export_models_url"
    template_name = "edc_export/export_models.html"

    def get_context_data(self, **kwargs) -> dict:
        if self.request.session.get("selected_models"):
            kwargs.update(
                selected_models=[
                    ModelOptions(**dct) for dct in self.request.session["selected_models"]
                ]
            )
        return super().get_context_data(**kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponseRedirect:  # noqa: ARG002
        if not self.check_user(request):
            pass
        elif not request.user.email:
            user_url = reverse("admin:auth_user_change", args=(request.user.id,))
            messages.error(
                request,
                format_html(
                    "Your account does not include an email address. "
                    'Please update your <a href="{user_url}">user account</a> '
                    "and try again.",
                    user_url=mark_safe(user_url),  # nosec B308 B703
                ),
            )
        else:
            selected_models = self.check_export_permissions(
                self.get_selected_models_from_session(request)
                or self.get_selected_models_from_post(request)
            )
            selected_models = [x.label_lower for x in selected_models]
            if not selected_models:
                messages.warning(
                    request,
                    "Nothing to do. Select one or more models and try again.",
                )
            else:
                try:
                    models_to_file = ModelsToFile(
                        models=selected_models,
                        user=request.user,
                        archive_to_single_file=True,
                        export_format=CSV,
                    )
                except ModelsToFileNothingExportedError as e:
                    messages.warning(request, f"Nothing to do. {e}.")
                else:
                    if settings.DEBUG:
                        messages.success(
                            request,
                            (
                                "Your data request has been saved to "
                                f"{models_to_file.archive_filename}. settings.DEBUG=True."
                            ),
                        )
                        update_data_request_history(request, models_to_file)
                    else:
                        try:
                            email_files_to_user(request, models_to_file)
                        except FilesEmailerError as e:
                            messages.error(
                                request, f"Failed to send the data you requested. Got '{e}'"
                            )
                        update_data_request_history(request, models_to_file)

        url = reverse(self.post_action_url, kwargs=self.kwargs)
        return HttpResponseRedirect(url)

    @staticmethod
    def check_export_permissions(selected_models) -> list[ModelOptions]:
        return selected_models

    @staticmethod
    def get_selected_models_from_post(request: WSGIRequest) -> list[ModelOptions]:
        """Returns a list of selected models from the POST
        as ModelOptions.
        """
        exportables = ExportableModelsForUser(request=request, user=request.user)
        selected_models = []
        for exportable in exportables:
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_models") or [])
            selected_models.extend(
                request.POST.getlist(f"chk_{exportable}_historical_models") or []
            )
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_list_models") or [])
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_inlines") or [])
        if selected_models:
            selected_models = [ModelOptions(model=m) for m in selected_models if m]
        return selected_models

    @staticmethod
    def get_selected_models_from_session(request: WSGIRequest) -> list[ModelOptions]:
        """Returns a list of selected models from the session object
        as ModelOptions.
        """
        try:
            selected_models = request.session.pop("selected_models")
        except KeyError:
            selected_models = []
        else:
            selected_models = [ModelOptions(**dct) for dct in selected_models]
        return selected_models

    def check_user(self, request) -> bool:
        try:
            valid_user = User.objects.filter(username=self.request.user).exists()
        except ObjectDoesNotExist:
            messages.error(request, "Invalid user.")
            valid_user = False
        return valid_user



# Generated by Django 4.2.1 on 2023-07-07 19:32

from django.db import migrations

import edc_identifier.managers
import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_unblinding", "0008_auto_20220826_0406"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="unblindingrequest",
            managers=[
                ("objects", edc_identifier.managers.SubjectIdentifierManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="unblindingreview",
            managers=[
                ("objects", edc_identifier.managers.SubjectIdentifierManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist

from edc_consent.modeladmin_mixins import RequiresConsentModelAdminMixin
from edc_fieldsets.fieldsets_modeladmin_mixin import FieldsetsModelAdminMixin
from edc_visit_tracking.modeladmin_mixins import CrfModelAdminMixin

from .model_admin_subject_dashboard_mixin import ModelAdminSubjectDashboardMixin


class ModelAdminCrfDashboardMixin(
    FieldsetsModelAdminMixin,
    RequiresConsentModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    CrfModelAdminMixin,
):
    show_save_next = True
    show_cancel = True
    show_dashboard_in_list_display_pos = 1

    def get_subject_dashboard_url_kwargs(self, obj) -> dict:
        return dict(
            subject_identifier=obj.related_visit.subject_identifier,
            appointment=str(obj.related_visit.appointment.id),
        )

    def get_changeform_initial_data(self, request) -> dict:
        initial_data = super().get_changeform_initial_data(request)
        try:
            related_visit = self.related_visit(request)
        except ObjectDoesNotExist:
            report_datetime = None
        else:
            report_datetime = getattr(related_visit, self.report_datetime_field_attr, None)
        initial_data.update(report_datetime=report_datetime)
        return initial_data



from __future__ import annotations

from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import PROTECT
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from .location import Location


class StorageBin(SiteModelMixin, BaseUuidModel):
    bin_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    bin_datetime = models.DateTimeField(default=timezone.now)

    name = models.CharField(
        max_length=10,
        null=True,
        blank=True,
        unique=True,
        help_text="May be left blank or set to any unique name.",
    )

    location = models.ForeignKey(Location, on_delete=PROTECT, null=True, blank=False)

    capacity = models.PositiveIntegerField(default=0, validators=[MinValueValidator(0)])

    in_use = models.BooleanField(default=True)

    description = models.CharField(max_length=100, default="", blank=True)

    history = HistoricalRecords()

    def __str__(self):
        location = self.location.site.id if self.location.site else self.location.name
        return f"{location}:{self.bin_identifier}"

    def save(self, *args, **kwargs):
        self.site = self.location.site
        if not self.id:
            self.bin_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if not self.name:
            self.name = self.bin_identifier
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Storage bin"
        verbose_name_plural = "Storage bins"



# Generated by Django 3.0.4 on 2020-05-12 01:52

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_export", "0006_auto_20200512_0208"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="plan",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
            },
        ),
    ]



from __future__ import annotations

from datetime import date
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from clinicedc_constants import DEAD, DEATH_REPORT_NOT_FOUND
from django import forms
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from edc_utils.text import convert_php_dateformat

from ..utils import get_ae_model

if TYPE_CHECKING:
    from ..model_mixins import DeathReportModelMixin


class BaseRequiresDeathReportFormValidatorMixin:
    @property
    def subject_identifier(self) -> str:
        return self.cleaned_data.get("subject_identifier") or self.instance.subject_identifier

    @property
    def death_report_model_cls(self) -> DeathReportModelMixin:
        return get_ae_model("deathreport")

    @property
    def death_report(self) -> DeathReportModelMixin | None:
        """Returns a model instance, if found, or None, if not."""
        try:
            return self.death_report_model_cls.objects.get(
                subject_identifier=self.subject_identifier
            )
        except ObjectDoesNotExist:
            return None

    @property
    def death_report_or_raises(self) -> DeathReportModelMixin:
        """Returns a model instance or raises
        forms.ValidationError.
        """
        try:
            return self.death_report_model_cls.objects.get(
                subject_identifier=self.subject_identifier
            )
        except ObjectDoesNotExist as e:
            verbose_name = self.death_report_model_cls._meta.verbose_name
            self.raise_validation_error(
                f"`{verbose_name}` not found.", DEATH_REPORT_NOT_FOUND, exc=e
            )

    @property
    def death_report_date(self) -> date:
        """Returns the localized death date from the death report
        instance.
        """
        death_report = self.death_report_or_raises
        value = getattr(death_report, death_report.death_date_field)

        try:
            death_report_date = value.astimezone(ZoneInfo("UTC")).date()
        except AttributeError:
            death_report_date = value
        return death_report_date


class RequiresDeathReportFormValidatorMixin(BaseRequiresDeathReportFormValidatorMixin):
    """A form validator mixin used by forms that refer to the
    death report.

    For example: off study report, study termination, etc.

        class StudyTerminationFormValidator(
            DeathReportFormValidatorMixin, FormValidator):

            def clean(self):

                self.validate_death_report_if_deceased()
                ...
    """

    offschedule_reason_field = "termination_reason"
    death_date_field = "death_date"  # on this form, for example, offschedule

    def validate_death_report_if_deceased(self) -> None:
        """Validates death report exists of termination_reason
        is "DEAD.

        Death "date" is the naive date of the settings.TIME_ZONE
        datetime.

        Note: uses __date field lookup. If using mysql don't forget
        to load timezone info.
        """

        if self.cleaned_data.get(self.offschedule_reason_field):
            if (
                self.cleaned_data.get(self.offschedule_reason_field).name == DEAD
                and not self.death_report
            ):
                raise forms.ValidationError(
                    {
                        self.offschedule_reason_field: "Patient is deceased, please complete "
                        "death report form first."
                    }
                )
            if (
                self.cleaned_data.get(self.offschedule_reason_field).name != DEAD
                and self.death_report
            ):
                raise forms.ValidationError(
                    {
                        self.offschedule_reason_field: (
                            "Invalid selection. A death report was submitted"
                        )
                    }
                )

        if not self.cleaned_data.get(self.death_date_field) and self.death_report:
            raise forms.ValidationError(
                {
                    self.death_date_field: (
                        "This field is required. A death report was submitted."
                    )
                }
            )
        if self.cleaned_data.get(self.death_date_field) and self.death_report:
            self.match_date_of_death_or_raise()

    def match_date_of_death_or_raise(self) -> None:
        """Raises an exception if the death date reported here does
        not match that from the Death Report.
        """
        if death_date := self.cleaned_data.get(self.death_date_field):
            try:
                death_date = death_date.date()
            except AttributeError:
                pass
            if self.death_report_date and self.death_report_date != death_date:
                expected = self.death_report_date.strftime(
                    convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                )
                actual = death_date.strftime(
                    convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                )
                verbose_name = self.death_report_model_cls._meta.verbose_name
                raise forms.ValidationError(
                    {
                        self.death_date_field: (
                            f"Date does not match `{verbose_name}`. "
                            f"Expected {expected}. Got {actual}."
                        )
                    }
                )



from clinicedc_constants import (
    ALREADY_REPORTED,
    GRADE3,
    GRADE4,
    NO,
    NOT_APPLICABLE,
    PRESENT_AT_BASELINE,
)

REPORTABLE = (
    (NOT_APPLICABLE, "Not applicable"),
    (GRADE3, "Yes, grade 3"),
    (GRADE4, "Yes, grade 4"),
    (NO, "Not reportable"),
    (ALREADY_REPORTED, "Already reported"),
    (PRESENT_AT_BASELINE, "Present at baseline"),
)



import random
import re

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from .checkdigit_mixins import LuhnOrdMixin
from .models import IdentifierModel


class ShortIdentifierError(Exception):
    pass


class ShortIdentifierPrefixError(Exception):
    pass


class ShortIdentifierPrefixPatternError(Exception):
    pass


class DuplicateIdentifierError(Exception):
    pass


class ShortIdentifier:
    name = "shortidentifier"
    template = "{prefix}{random_string}"
    random_string_pattern = r"[A-Z0-9]+"  # alhpanumeric
    random_string_length = 5
    prefix_pattern = r"^[0-9]{2}$"
    identifier_model_cls = IdentifierModel

    prefix = None
    seed = None

    checkdigit = LuhnOrdMixin()

    def __init__(
        self,
        name=None,
        prefix_pattern=None,
        prefix=None,
        template=None,
        random_string_length=None,
        random_string_pattern=None,
    ):
        self._last_random_string = None
        self.name = name or self.name
        self.template = template or self.template
        self.random_string_length = random_string_length or self.random_string_length
        self.random_string_pattern = random_string_pattern or self.random_string_pattern
        self.random_string_pattern = re.compile(self.random_string_pattern)
        self.prefix_pattern = prefix_pattern or self.prefix_pattern
        self.prefix = prefix or self.prefix or ""
        self.prefix = str(self.prefix)
        if self.prefix and not self.prefix_pattern:
            raise ShortIdentifierPrefixError(
                "Prefix declared without a corresponding pattern. "
                f"Got prefix='{self.prefix}', "
                f"prefix_pattern='{self.prefix_pattern}'."
            )

        if self.prefix_pattern:
            if not self.prefix_pattern.startswith("^") or not self.prefix_pattern.endswith(
                "$"
            ):
                raise ShortIdentifierPrefixPatternError(
                    f"Invalid prefix pattern. Got {self.prefix_pattern}."
                )
            self.prefix_pattern = re.compile(self.prefix_pattern)

        if not self.prefix and self.prefix_pattern:
            raise ShortIdentifierPrefixError(
                "Prefix does not match prefix pattern. Got prefix=None."
            )
        if self.prefix and not self.prefix_pattern.match(self.prefix):
            raise ShortIdentifierPrefixError(
                "Prefix does not match prefix pattern. "
                f"Got '{self.prefix}' does not match "
                f"pattern '{self.prefix_pattern}'."
            )

        edc_device_app_config = django_apps.get_app_config("edc_device")
        self.device_id = edc_device_app_config.device_id

        self.identifier = self.get_identifier()

    def __str__(self):
        return self.identifier

    def get_identifier(self):
        """Returns a new unique identifier."""
        identifier = None
        allowed_chars = self.random_string_pattern.match("ABCDEFGHKMNPRTUVWXYZ2346789").group()
        max_tries = len(allowed_chars) ** (self.random_string_length + 1)
        tries = 0
        while not identifier:
            tries += 1
            random_string = "".join(
                [
                    random.choice(allowed_chars)  # nosec B311  # noqa: S311
                    for _ in range(self.random_string_length)
                ]
            )
            identifier = self.template.format(random_string=random_string, prefix=self.prefix)
            try:
                self.identifier_model_cls.objects.get(
                    identifier=identifier, identifier_type=self.name
                )
            except ObjectDoesNotExist:
                pass
            else:
                identifier = None
                if tries >= max_tries:
                    raise DuplicateIdentifierError(
                        "Unable prepare a unique requisition identifier, "
                        "all are taken. Increase the length of the random string. "
                        f"tries={tries}, max_tries={max_tries}."
                    )

        self.identifier_model_cls.objects.create(
            identifier=identifier,
            identifier_type=self.name,
            identifier_prefix=self.prefix,
            device_id=self.device_id,
        )
        return identifier



# Generated by Django 5.1.2 on 2024-11-17 16:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0034_rename_display_label_assignment_display_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="container",
            name="max_per_subject",
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True),
        ),
        migrations.AddField(
            model_name="historicalcontainer",
            name="max_per_subject",
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True),
        ),
    ]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from clinicedc_constants import NO
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.utils.translation import gettext_lazy as _

from edc_form_validators import INVALID_ERROR, FormValidator

from ..constants import NEW_APPT
from ..utils import get_allow_skipped_appt_using, validate_date_is_on_clinic_day

if TYPE_CHECKING:
    from edc_facility.models import HealthFacility


class NextAppointmentCrfFormValidatorMixin(FormValidator):
    def __init__(self, **kwargs):
        self._clinic_days = None
        self._health_facility = None
        super().__init__(**kwargs)

    def clean(self):
        self.required_if(
            NO,
            field="offschedule_today",
            field_required="appt_date",
        )
        self.required_if(
            NO,
            field="offschedule_today",
            field_required="visitschedule",
        )

        self.required_if(
            NO,
            field="offschedule_today",
            field_required="health_facility",
        )

        self.required_if(
            NO,
            field="offschedule_today",
            field_required="health_facility",
        )

        if self.cleaned_data.get("appt_date") and self.cleaned_data.get("visitschedule"):
            self.validate_with_appointment_form_validator()
            self.validate_date_is_on_clinic_day()

    @property
    def visit_code_fld(self):
        if get_allow_skipped_appt_using():
            return get_allow_skipped_appt_using().get(self.model._meta.label_lower)[1]
        return "visitschedule"

    @property
    def visit_code(self):
        return getattr(
            self.cleaned_data.get(self.visit_code_fld),
            "visit_code",
            self.cleaned_data.get(self.visit_code_fld),
        )

    def validate_with_appointment_form_validator(self):
        # use AppointmentFormValidator to validate the next appt data
        from ..form_validators import AppointmentFormValidator as Base
        from ..models import Appointment

        class AppointmentFormValidator(Base):
            def validate_not_future_appt_datetime(self):
                pass

        try:
            instance = Appointment.objects.get(
                subject_identifier=self.related_visit.subject_identifier,
                visit_schedule_name=self.related_visit.visit_schedule_name,
                schedule_name=self.related_visit.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=0,
            )
        except ObjectDoesNotExist as e:
            self.raise_validation_error(
                {
                    self.visit_code_fld: (
                        "Invalid selection. Expected "
                        f"{getattr(self.related_visit.appointment.next, 'visit_code', 'None')}"
                    )
                },
                INVALID_ERROR,
                exc=e,
            )
        if instance == self.related_visit.appointment:
            self.raise_validation_error(
                {self.visit_code_fld: "Cannot be the current visit"}, INVALID_ERROR
            )

        if (
            instance.appt_status != NEW_APPT
            and instance.appt_datetime.date() != self.cleaned_data.get("appt_date")
        ):
            self.raise_validation_error(
                {"appt_date": "May not be changed. Next appointment has already started"},
                INVALID_ERROR,
            )

        cleaned_data = instance.__dict__
        cleaned_data.update(appt_datetime=self.appt_datetime)
        appointment_validator = AppointmentFormValidator(
            cleaned_data=cleaned_data,
            instance=instance,
            model=Appointment,
        )
        try:
            appointment_validator.validate()
        except ValidationError as e:
            if e.message_dict.get("appt_datetime"):
                raise ValidationError({"appt_date": e.message_dict["appt_datetime"]}) from e
            raise

    @property
    def appt_datetime(self) -> datetime:
        return datetime(
            self.cleaned_data.get("appt_date").year,
            self.cleaned_data.get("appt_date").month,
            self.cleaned_data.get("appt_date").day,
            7,
            30,
            0,
            tzinfo=ZoneInfo("UTC"),
        )

    @property
    def clinic_days(self) -> list[int]:
        if not self._clinic_days and self.cleaned_data.get("health_facility"):
            self._clinic_days = self.health_facility.clinic_days
        return self._clinic_days

    def validate_date_is_on_clinic_day(self):
        validate_date_is_on_clinic_day(
            cleaned_data=dict(
                appt_date=self.cleaned_data.get("appt_date"),
                report_datetime=self.cleaned_data.get("report_datetime"),
            ),
            clinic_days=self.clinic_days,
            raise_validation_error=self.raise_validation_error,
        )

    @property
    def health_facility(self) -> HealthFacility | None:
        if not self._health_facility:
            if self.cleaned_data.get("health_facility"):
                if self.cleaned_data.get("health_facility").site != self.related_visit.site:
                    raise self.raise_validation_error(
                        {"health_facility": "Invalid for this site"}, INVALID_ERROR
                    )
            else:
                raise self.raise_validation_error(
                    {"health_facility": _("This field is required.")}, INVALID_ERROR
                )
            self._health_facility = self.cleaned_data.get("health_facility")
        return self._health_facility



from django.db import models
from django.db.models import PROTECT
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords

from ..medication import Assignment, Formulation


class Manager(models.Manager):
    use_in_migrations = True


class Product(BaseUuidModel):

    product_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    name = models.CharField(max_length=50, blank=True, help_text="Leave blank to use default")

    formulation = models.ForeignKey(Formulation, on_delete=PROTECT)

    assignment = models.ForeignKey(Assignment, on_delete=PROTECT, null=True, blank=False)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        """Note: For unblinded users, add assignment in ModelAdmin where a request
        object is available.
        """
        return self.formulation.get_description_with_assignment(self.assignment)

    def save(self, *args, **kwargs):
        if not self.product_identifier:
            self.product_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if not self.name:
            self.name = self.formulation.get_product_description()
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Product"
        verbose_name_plural = "Product"



from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from ..constants import PROTOCOL_DEVIATION_VIOLATION_ACTION
from ..model_mixins import ProtocolDeviationViolationModelMixin


class ProtocolDeviationViolation(
    ProtocolDeviationViolationModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    ActionModelMixin,
    BaseUuidModel,
):
    action_name = PROTOCOL_DEVIATION_VIOLATION_ACTION
    on_site = ActionIdentifierSiteManager()
    objects = ActionIdentifierModelManager()

    def natural_key(self):
        return (self.action_identifier,)

    class Meta(ProtocolDeviationViolationModelMixin.Meta, BaseUuidModel.Meta):
        indexes = (
            *ProtocolDeviationViolationModelMixin.Meta.indexes,
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            *ActionModelMixin.Meta.indexes,
            *BaseUuidModel.Meta.indexes,
        )



from django.contrib.sites.models import Site
from django.db import models
from django.db.models import PROTECT, Index, UniqueConstraint

from edc_model.models import BaseUuidModel


class Issue(BaseUuidModel):
    session_id = models.UUIDField(null=True)
    session_datetime = models.DateTimeField(null=True)
    label_lower = models.CharField(max_length=50)
    verbose_name = models.CharField(max_length=150)
    subject_identifier = models.CharField(max_length=50)
    visit_code = models.CharField(max_length=25, null=True)
    visit_code_sequence = models.IntegerField(null=True)
    visit_schedule_name = models.CharField(max_length=25, null=True)
    schedule_name = models.CharField(max_length=25, null=True)
    field_name = models.CharField(max_length=150)
    raw_message = models.TextField(null=True)
    message = models.TextField(null=True)
    short_message = models.CharField(max_length=250, null=True)
    response = models.CharField(max_length=250, null=True)
    src_id = models.UUIDField(null=True)
    src_revision = models.CharField(max_length=150, null=True)
    src_report_datetime = models.DateTimeField(null=True)
    src_created_datetime = models.DateTimeField(null=True)
    src_modified_datetime = models.DateTimeField(null=True)
    src_user_created = models.CharField(max_length=150, null=True)
    src_user_modified = models.CharField(max_length=150, null=True)
    site = models.ForeignKey(Site, on_delete=PROTECT)
    panel_name = models.CharField(max_length=50, null=True)
    extra_formfields = models.TextField(null=True)
    exclude_formfields = models.TextField(null=True)

    def __str__(self):
        return (
            f"{self.subject_identifier} {self.visit_code}.{self.visit_code_sequence} "
            f"{self.verbose_name} ({self.src_revision.split(':')[0]}):{self.field_name} "
            f"{self.message}."
        )

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Issue"
        verbose_name_plural = "Issues"
        constraints = [
            UniqueConstraint(
                fields=[
                    "subject_identifier",
                    "label_lower",
                    "panel_name",
                    "visit_code",
                    "visit_code_sequence",
                    "visit_schedule_name",
                    "schedule_name",
                    "field_name",
                ],
                name="unique_label_lower_subject_identifier_etc",
            )
        ]
        indexes = (
            *BaseUuidModel.Meta.indexes,
            Index(fields=["label_lower", "field_name", "panel_name", "short_message"]),
            Index(
                fields=[
                    "subject_identifier",
                    "visit_code",
                    "visit_code_sequence",
                    "label_lower",
                    "panel_name",
                    "short_message",
                ],
            ),
        )



# Generated by Django 4.2.7 on 2023-12-04 02:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_crf", "0005_alter_crfstatus_options_crfstatus_locale_created_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="crfstatus",
            name="edc_crf_crf_modifie_6152c9_idx",
        ),
        migrations.AddIndex(
            model_name="crfstatus",
            index=models.Index(
                fields=["modified", "created"], name="edc_crf_crf_modifie_709d43_idx"
            ),
        ),
    ]
